// -*- mode: c++; c-file-style: "k&r"; c-basic-offset: 4 -*-
/***********************************************************************
 *
 * vr/replica.cc:
 *   Viewstamped Replication protocol
 *
 * Copyright 2013-2016 Dan R. K. Ports  <drkp@cs.washington.edu>
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 **********************************************************************/

#include "common/replica.h"
#include "vr/replica.h"
#include "vr/vr-proto.pb.h"

#include "lib/assert.h"
#include "lib/configuration.h"
#include "lib/latency.h"
#include "lib/message.h"
#include "lib/transport.h"
#include "lib/udptransport.h"

#include <algorithm>
#include <random>
#include <string>

#ifdef __cplusplus
extern "C"
{
#endif

#include <asm-generic/posix_types.h>
#include <linux/if_link.h>
#include <linux/limits.h>

#include <linux/bpf.h>
#include <bpf/bpf.h>
#include <bpf/libbpf.h>
#include <xdp-handler/fast_common.h>

#ifdef __cplusplus
}
#endif

// #define FAST_QUORUM_PRUNE
// #define FAST_BATCH // unused!
// #define FAST_REPLY
// #define TC_BROADCAST

#define RDebug(fmt, ...) Debug("[%d] " fmt, myIdx, ##__VA_ARGS__)
#define RNotice(fmt, ...) Notice("[%d] " fmt, myIdx, ##__VA_ARGS__)
#define RWarning(fmt, ...) Warning("[%d] " fmt, myIdx, ##__VA_ARGS__)
#define RPanic(fmt, ...) Panic("[%d] " fmt, myIdx, ##__VA_ARGS__)

namespace specpaxos
{
    namespace vr
    {

        using namespace proto;

#ifdef FAST_BATCH                                                  // 预处理指令，预处理指令
        int request_polling(void *ctx, void *data, size_t data_sz) // 函数接受三个参数：ctx（上下文指针），data（数据指针），data_sz（数据大小） by：wbh
        {

            VRReplica *pt = (VRReplica *)ctx;
            // 将 ctx 转换为 VRReplica 指针类型

            sockaddr_in sender;
            sender.sin_family = AF_INET;
            sender.sin_port = *(in_port_t *)data;
            data += sizeof(in_port_t);
            sender.sin_addr.s_addr = *(in_addr_t *)data;
            data += sizeof(in_addr_t);
            // 从 data 中解析出发送者的信息，包括端口号和 IP 地址，并将其保存在 sender 结构体中。

            uint64_t size = *(uint64_t *)data;
            static RequestMessage request;
            request.ParseFromString(string((char *)data + sizeof(uint64_t), size));

            UDPTransportAddress senderAddr(sender);
            pt->HandleRequest(senderAddr, request);
            // 创建一个 UDPTransportAddress 对象，表示发送者的地址，并调用 HandleRequest 方法处理请求

            return 0;
        }
#endif

#ifdef FAST_REPLY
        int prepare_polling(void *ctx, void *data, size_t data_sz)
        {
            VRReplica *pt = (VRReplica *)ctx;
            // 将 ctx 转换为 VRReplica 指针类型 by：wbh
            uint64_t size = *(uint64_t *)data;
            static PrepareMessage prepare;
            prepare.ParseFromString(string((char *)data + sizeof(uint64_t), size));
            // 创建了一个 PrepareMessage 类型的静态变量 prepare，然后调用 prepare.ParseFromString 方法将数据解析到 prepare 变量中

            pt->HandlePrepare_Kernel(prepare);
            return 0;
        }
#endif

        VRReplica::VRReplica(Configuration config, int myIdx,
                             bool initialize,
                             Transport *transport, int batchSize,
                             AppReplica *app)
            : Replica(config, myIdx, initialize, transport, app),
              batchSize(batchSize),
              log(false),
              prepareOKQuorum(config.QuorumSize() - 1), // set the quorum.
              startViewChangeQuorum(config.QuorumSize() - 1),
              doViewChangeQuorum(config.QuorumSize() - 1), // quorum of view change, basically the same.
              recoveryResponseQuorum(config.QuorumSize())
        {
            this->status = STATUS_NORMAL; // State is normal.
            this->view = 0;
            this->lastOp = 0;
            this->lastCommitted = 0;
            this->lastRequestStateTransferView = 0;
            this->lastRequestStateTransferOpnum = 0;
            lastBatchEnd = 0;
            batchComplete = true;

            if (batchSize > 1)
            {
                Notice("Batching enabled; batch size %d", batchSize);
            }

            // If we don't hear from leader for a period of time, then do the view change.
            this->viewChangeTimeout = new Timeout(transport, 5000, [this, myIdx]()
                                                  {
            RWarning("Have not heard from leader; starting view change");
            StartViewChange(view+1); });
            this->nullCommitTimeout = new Timeout(transport, 1000, [this]()
                                                  { SendNullCommit(); });
            this->stateTransferTimeout = new Timeout(transport, 1000, [this]()
                                                     {
            this->lastRequestStateTransferView = 0;
            this->lastRequestStateTransferOpnum = 0; });
            this->stateTransferTimeout->Start();
            this->resendPrepareTimeout = new Timeout(transport, 500, [this]()
                                                     { ResendPrepare(); });
            this->closeBatchTimeout = new Timeout(transport, 300, [this]()
            // 存在FAST_BATCH就多调用ring_buffer_poll       {

#ifdef FAST_BATCH
            ring_buffer__poll(rb_request, 0);
#endif
            CloseBatch();
        });
        this->recoveryTimeout = new Timeout(transport, 5000, [this]()
                                            { SendRecoveryMessages(); });

        _Latency_Init(&requestLatency, "request");
        _Latency_Init(&executeAndReplyLatency, "executeAndReply");

        if (initialize)
        {
            if (AmLeader())
            {
                nullCommitTimeout->Start(); // start heart-beating message.
            }
            else
            {
                viewChangeTimeout->Start();
            }
        }
        else
        { // recovery... this should be rare case.
            this->status = STATUS_RECOVERING;
            this->recoveryNonce = GenerateNonce();
            SendRecoveryMessages();
            recoveryTimeout->Start();
        }

        memset(sgn_bits, 0, sizeof(sgn_bits));
        sgn_bits[0] = (1 << 31); // #define BROADCAST_SIGN_BIT (1<<31)
                                 // 用于消息广播的一个接口？在发消息的函数中新增调用    by：wbh

        // bpf_obj_get：通过文件名称获取本进程中的句柄fd    by：wbh
        // 获得了一系列需要使用的句柄
#ifdef FAST_BATCH
        request_buffer_fd = bpf_obj_get("/sys/fs/bpf/paxos_request_buffer");
        if (request_buffer_fd < 0)
        {
            fprintf(stderr, "Error: bpf_object__find_map_fd_by_name \"paxos_request_buffer\" failed\n");
            exit(1); // return 1;
        }
        rb_request = ring_buffer__new(request_buffer_fd, request_polling, this, NULL);
        if (!rb_request)
        {
            fprintf(stderr, "Failed to create request ring buffer\n");
            exit(-1);
        }
#endif
#ifdef FAST_REPLY
        prepare_buffer_fd = bpf_obj_get("/sys/fs/bpf/paxos_prepare_buffer");
        if (prepare_buffer_fd < 0)
        {
            fprintf(stderr, "Error: bpf_object__find_map_fd_by_name \"paxos_prepare_buffer\" failed\n");
            exit(1); // return 1;
        }
        rb_prepare = ring_buffer__new(prepare_buffer_fd, prepare_polling, this, NULL);
        if (!rb_prepare)
        {
            fprintf(stderr, "Failed to create prepare ring buffer\n");
            exit(-1);
        }
#endif

#if defined FAST_REPLY || defined FAST_BATCH || defined FAST_QUORUM_PRUNE
        paxos_ctr_state_fd = bpf_obj_get("/sys/fs/bpf/paxos_ctr_state");
        if (paxos_ctr_state_fd < 0)
        {
            fprintf(stderr, "Error: bpf_object__find_map_fd_by_name \"paxos_ctr_state\" failed\n");
            exit(1); // return 1;
        }
        ModifyKernelState();
#endif
    }

    VRReplica::~VRReplica()
    {
        Latency_Dump(&requestLatency);
        Latency_Dump(&executeAndReplyLatency);

        delete viewChangeTimeout;
        delete nullCommitTimeout;
        delete stateTransferTimeout;
        delete resendPrepareTimeout;
        delete closeBatchTimeout;
        delete recoveryTimeout;

        for (auto &kv : pendingPrepares)
        {
            delete kv.first;
        }
    }

#if defined FAST_REPLY || defined FAST_BATCH || defined FAST_QUORUM_PRUNE
    void VRReplica::ModifyKernelState() // 修改内核状态 by：wbh
    {
        // 定义了一个结构体paxos_ctr_state，包含了一些状态信息
        struct paxos_ctr_state
        {
            enum ReplicaStatus state;
            int myIdx, leaderIdx, batchSize;
            __u64 view, lastOp;
        } state;

        state.state = status;                                 // 当前状态
        state.view = view;                                    // 当前视图号(？
        state.lastOp = lastOp;                                // 最后一次操作
        state.myIdx = myIdx;                                  // 当前节点的索引
        state.leaderIdx = configuration.GetLeaderIndex(view); // 领导者节点的索引
        state.batchSize = batchSize;                          // 批处理大小

        unsigned int zero = 0;
        // 调用bpf_map_update_elem函数将其更新到名为paxos_ctr_state_fd的BPF映射
        if (bpf_map_update_elem(paxos_ctr_state_fd, &zero, &state, 0) < 0)
        {
            fprintf(stderr, "Error: bpf_map_update_elem \"ctr_state\" failed\n");
            exit(1);
        }
        // asd123www: maybe here is a polling?
    }
#endif
    // 生成一个随机数
    uint64_t
    VRReplica::GenerateNonce() const
    {
        std::random_device rd;
        std::mt19937_64 gen(rd());
        std::uniform_int_distribution<uint64_t> dis;
        return dis(gen);
    }

    bool
    VRReplica::AmLeader() const
    {
        return (configuration.GetLeaderIndex(view) == myIdx);
    }
    /**
     * VRReplica::CommitUpTo
     *
     * 提交日志中的操作，直到指定的操作号。
     * 这个过程包括执行操作、更新状态、发送回复等。
     *
     * @param upto 指定需要提交到哪个操作号。
     */
    void
    VRReplica::CommitUpTo(opnum_t upto)
    {
        // 循环提交日志中的操作，直到达到指定的操作号
        while (lastCommitted < upto)
        {
            Latency_Start(&executeAndReplyLatency); // 开始测量执行和回复的延迟

            lastCommitted++; // 提交编号递增

            // 在日志中查找当前操作
            const LogEntry *entry = log.Find(lastCommitted);
            if (!entry)
            {
                RPanic("Did not find operation " FMT_OPNUM " in log", lastCommitted);
            }

            // 执行找到的操作
            RDebug("Executing request " FMT_OPNUM, lastCommitted);
            ReplyMessage reply;
            Execute(lastCommitted, entry->request, reply);

            // 设置回复中的视图和操作号
            reply.set_view(entry->viewstamp.view);
            reply.set_opnum(entry->viewstamp.opnum);
            reply.set_clientreqid(entry->request.clientreqid());

            // 标记操作为已提交
            log.SetStatus(lastCommitted, LOG_STATE_COMMITTED);

            // 更新客户端表，记录回复
            ClientTableEntry &cte =
                clientTable[entry->request.clientid()];
            if (cte.lastReqId <= entry->request.clientreqid())
            {
                cte.lastReqId = entry->request.clientreqid();
                cte.replied = true;
                cte.reply = reply;
            }
            else
            {
                // 如果有后续的操作被准备，那么这个请求被认为已在客户端完成，不需要记录结果
            }

            // 如果找到客户端的地址，发送回复
            auto iter = clientAddresses.find(entry->request.clientid());
            if (iter != clientAddresses.end())
            {
                transport->SendMessage(this, *iter->second, reply);
            }

            Latency_End(&executeAndReplyLatency); // 结束测量
        }
    }

    /**
     * @brief 向其他副本发送PREPAREOK消息，用于确认新的未提交操作。
     *
     * 这个函数遍历从oldLastOp到lastOp的所有操作，对那些新且未提交的操作，
     * 发送PREPAREOK消息。PREPAREOK消息包含操作号、视图号和发送者的索引。
     *
     * @param oldLastOp 旧的最后一个操作号，表示从此操作号开始检查和发送PREPAREOK消息。
     */
    void
    VRReplica::SendPrepareOKs(opnum_t oldLastOp)
    {
        // 遍历日志中的操作，寻找需要发送PREPAREOK的消息的操作
        for (opnum_t i = oldLastOp; i <= lastOp; i++)
        {
            // 如果操作已经被提交，则跳过
            if (i <= lastCommitted)
            {
                continue;
            }

            // 在日志中查找当前操作
            const LogEntry *entry = log.Find(i);
            // 如果找不到操作，则panic
            if (!entry)
            {
                RPanic("Did not find operation " FMT_OPNUM " in log", i);
            }
            // 断言：找到的操作必须处于PREPARED状态
            ASSERT(entry->state == LOG_STATE_PREPARED);
            // 更新客户端表
            UpdateClientTable(entry->request);

            // 准备PREPAREOK消息
            PrepareOKMessage reply;
            reply.set_view(view);
            reply.set_opnum(i);
            reply.set_replicaidx(myIdx);

            // 打印调试信息，说明正在发送PREPAREOK消息
            uint32_t my_data[3] = {(uint32_t)view, (uint32_t)i, (uint32_t)myIdx};
            RDebug("Sending PREPAREOK " FMT_VIEWSTAMP " for new uncommitted operation",
                   reply.view(), reply.opnum());

            // 发送PREPAREOK消息给领导者。如果发送失败，则打印警告信息。
            if (!(transport->SendMessageToReplica(this,
                                                  configuration.GetLeaderIndex(view),
                                                  reply, my_data)))
            {
                RWarning("Failed to send PrepareOK message to leader");
            }
        }
    }

    /**
     * 函数名称: SendRecoveryMessages
     * 功能描述: 发送恢复信息给所有副本，以请求恢复操作。
     * 方法参数: 无
     * 返回值: 无
     */
    void VRReplica::SendRecoveryMessages()
    {
        // 创建恢复消息并设置消息中的副本索引和随机数
        RecoveryMessage m;
        m.set_replicaidx(myIdx);
        m.set_nonce(recoveryNonce);

        // 输出日志，请求恢复
        RNotice("Requesting recovery");
        // 尝试通过传输层向所有副本发送恢复消息，若发送失败则记录警告信息
        if (!transport->(this, m))
        {
            RWarning("Failed to send Recovery message to all replicas");
        }
    }
    /**
     * 请求状态传输的函数。
     * 当发现当前状态陈旧时，会调用此函数来请求更新状态。
     * 该函数不会返回任何值。
     */
    void VRReplica::RequestStateTransfer()
    {
        // 创建状态传输请求消息并设置视图号和操作序号
        RequestStateTransferMessage m;
        m.set_view(view);
        m.set_opnum(lastCommitted);

        // 检查是否已经请求过相同的状态传输，如果是，则跳过此次请求
        if ((lastRequestStateTransferOpnum != 0) &&
            (lastRequestStateTransferView == view) &&
            (lastRequestStateTransferOpnum == lastCommitted))
        {
            RDebug("Skipping state transfer request " FMT_VIEWSTAMP
                   " because we already requested it",
                   view, lastCommitted);
            return;
        }

        // 记录并更新请求状态传输的视图号和操作序号
        RNotice("Requesting state transfer: " FMT_VIEWSTAMP, view, lastCommitted);
        this->lastRequestStateTransferView = view;
        this->lastRequestStateTransferOpnum = lastCommitted;

        // 向所有副本发送状态传输请求消息
        if (!transport->SendMessageToAll(this, m))
        {
            RWarning("Failed to send RequestStateTransfer message to all replicas");
        }
    }

    void
    VRReplica::EnterView(view_t newview)
    // 总的来说，该函数是用来进行视图切换操作的，根据当前节点是否为领导者进行不同的逻辑处理，
    // 并且在进入新视图时进行相应的状态更新和超时计时器管理。
    {
        RNotice("Entering new view " FMT_VIEW, newview);
        // 更新相关值
        view = newview;
        status = STATUS_NORMAL;
        lastBatchEnd = lastOp;
        batchComplete = true;

        recoveryTimeout->Stop();
        // 针对是否是领导节点、对一系列计时器进行开启和关闭操作
        if (AmLeader())
        {
            viewChangeTimeout->Stop();
            nullCommitTimeout->Start();
        }
        else
        {
            viewChangeTimeout->Start();
            nullCommitTimeout->Stop();
            resendPrepareTimeout->Stop();
            closeBatchTimeout->Stop();
        }

        prepareOKQuorum.Clear();
        startViewChangeQuorum.Clear();
        doViewChangeQuorum.Clear();
        recoveryResponseQuorum.Clear();
    }

    // if can't hear the heart beat from leader, then do the view change!
    void VRReplica::StartViewChange(view_t newview)
    {
        RNotice("Starting view change for view " FMT_VIEW, newview);

        view = newview;
        status = STATUS_VIEW_CHANGE;

        viewChangeTimeout->Reset();
        nullCommitTimeout->Stop();
        resendPrepareTimeout->Stop();
        closeBatchTimeout->Stop();

        StartViewChangeMessage m;
        m.set_view(newview);
        m.set_replicaidx(myIdx);
        m.set_lastcommitted(lastCommitted);
        // broadcast this viewchange message.
        if (!transport->SendMessageToAll(this, m))
        {
            RWarning("Failed to send StartViewChange message to all replicas");
        }
    }

    // heart beat message from leader.
    void VRReplica::SendNullCommit()
    {
        CommitMessage cm;
        cm.set_view(this->view);
        cm.set_opnum(this->lastCommitted);

        ASSERT(AmLeader());
        // 如果TC_BROADCAST存在的情况下，条件中多加一个sgn_bits     by：wbh
        // 修改了lib中transportcommon.h的SendMessageToAll函数
        // 简单查看中，sendmessage相关的几个函数都加了buf参数
        // 此处的sgn_bits是什么？  构造函数中定义的，目前理解是用于TC广播的一个接口？
        if (!(transport->SendMessageToAll(this, cm
#ifdef TC_BROADCAST
                                          ,
                                          sgn_bits
#endif
                                          )))
        {
            RWarning("Failed to send null COMMIT message to all replicas");
        }
    }

    // asd123www: this can be maintained in XDP.
    // 更新客户端表中的信息
    void VRReplica::UpdateClientTable(const Request &req)
    {
        ClientTableEntry &entry = clientTable[req.clientid()];

        ASSERT(entry.lastReqId <= req.clientreqid());

        if (entry.lastReqId == req.clientreqid())
        {
            return;
        }

        entry.lastReqId = req.clientreqid();
        entry.replied = false;
        entry.reply.Clear();
    }

    void VRReplica::ResendPrepare() //
    {
        ASSERT(AmLeader());
        if (lastOp == lastCommitted)
        {
            return;
        }
        RNotice("Resending prepare");
        // 同样的，涉及发送消息的函数在广播条件下加入一个sgn_bits参数
        if (!(transport->SendMessageToAll(this, lastPrepare
#ifdef TC_BROADCAST
                                          ,
                                          sgn_bits
#endif
                                          )))
        {
            RWarning("Failed to ressend prepare message to all replicas");
        }
    }

    void VRReplica::CloseBatch()
    {
        ASSERT(AmLeader());
        ASSERT(lastBatchEnd < lastOp); // the messages to be sent is not empty set.

        opnum_t batchStart = lastBatchEnd + 1;

        RDebug("Sending batched prepare from " FMT_OPNUM
               " to " FMT_OPNUM,
               batchStart, lastOp);
        /* Send prepare messages */
        // 设置准备信息
        PrepareMessage p;
        p.set_view(view);
        p.set_opnum(lastOp);
        p.set_batchstart(batchStart);

        // batch the reqs in this interval, and send it all.
        // 为每个操作编号查找对应的日志条目，并确保视图号和操作编号匹配，将请求添加到消息中。
        // by：wbh
        for (opnum_t i = batchStart; i <= lastOp; i++)
        {
            Request *r = p.add_request();
            const LogEntry *entry = log.Find(i);
            ASSERT(entry != NULL);
            ASSERT(entry->viewstamp.view == view);
            ASSERT(entry->viewstamp.opnum == i);
            *r = entry->request;
        }
        lastPrepare = p;

        // asd123www: raw data, easy to parse in XDP.
        // 原始数据，易于在XDP中解析，根据宏定义情况生成my_Data，这个是干什么用的？ 问题
        // my_buf 实际上的格式就是和my_data一样吗？
#ifndef TC_BROADCAST
        uint32_t my_data[3] = {(uint32_t)view, (uint32_t)lastOp, (uint32_t)batchStart};
#else
        uint32_t my_data[3] = {(uint32_t)view | (1 << 31), (uint32_t)lastOp, (uint32_t)batchStart};
#endif
        if (!(transport->SendMessageToAll(this, p, my_data)))
        {
            RWarning("Failed to send prepare message to all replicas");
        }
        // 更新最后批次的结束编号、将批次标记为未完成，重置重新发送 prepare 消息的计时器并停止关闭批处理的计时器
        lastBatchEnd = lastOp;
        batchComplete = false;

        resendPrepareTimeout->Reset();
        closeBatchTimeout->Stop();
    }

    void VRReplica::ReceiveMessage(const TransportAddress &remote, const string &type, const string &data)
    {
// 在#ifdef FAST_BATCH的条件下，如果当前节点是领导者，则使用assert断言来确保通过调用ring_buffer__poll函数来获取请求环形缓冲区中的请求数量不小于0。

// 在#ifdef FAST_REPLY的条件下，如果当前节点不是领导者，则使用assert断言来确保通过调用ring_buffer__poll函数来获取准备环形缓冲区中的请求数量不小于0
// by:wbh
#ifdef FAST_BATCH
        if (AmLeader())
        {
            assert(ring_buffer__poll(rb_request, 0) >= 0);
        }
#endif
#ifdef FAST_REPLY
        if (!AmLeader())
        {
            assert(ring_buffer__poll(rb_prepare, 0) >= 0);
        }
#endif

        static RequestMessage request;
        static UnloggedRequestMessage unloggedRequest;
        static PrepareMessage prepare;
        static PrepareOKMessage prepareOK;
        static CommitMessage commit;
        static RequestStateTransferMessage requestStateTransfer;
        static StateTransferMessage stateTransfer;
        static StartViewChangeMessage startViewChange;
        static DoViewChangeMessage doViewChange;
        static StartViewMessage startView;
        static RecoveryMessage recovery;
        static RecoveryResponseMessage recoveryResponse;
        static std::string my_PrepareOK("specpaxos.vr.MyPrepareOK");
        // 静态字符串变量my_PrepareOK的定义

        // 消息处理逻辑，根据收到的消息类型来调用相应的处理函数。
        if (type == request.GetTypeName())
        { // HandleRequest, the leader's duty.
            request.ParseFromString(data);
            HandleRequest(remote, request);
        }
        else if (type == unloggedRequest.GetTypeName())
        { // only by client, omit.
            unloggedRequest.ParseFromString(data);
            HandleUnloggedRequest(remote, unloggedRequest);
        }
        else if (type == prepare.GetTypeName())
        { // HandlePrepare, in backup replica.
            prepare.ParseFromString(data);
            HandlePrepare(remote, prepare);
        }
        else if (type == prepareOK.GetTypeName())
        { // HandlePrepareOK, the leader's duty.
            prepareOK.ParseFromString(data);
            HandlePrepareOK(remote, prepareOK);
        }
        else if (type == my_PrepareOK)
        { // our custom PrepareOK reply.
            // 针对自定义的消息类型my_PrepareOK的处理逻辑，解析数据，并调用HandlePrepareOK函数处理收到的PrepareOK消息。
            const char *c = data.c_str();
            prepareOK.set_view(*(uint64_t *)c);
            prepareOK.set_opnum(*((uint64_t *)c + 1));
            prepareOK.set_replicaidx(*((uint32_t *)c + 4));
            HandlePrepareOK(remote, prepareOK);
        }
        else if (type == commit.GetTypeName())
        { // HandleCommit, in back replica.
            commit.ParseFromString(data);
            HandleCommit(remote, commit);
        }
        else if (type == requestStateTransfer.GetTypeName())
        { // checked.
            requestStateTransfer.ParseFromString(data);
            HandleRequestStateTransfer(remote, requestStateTransfer);
        }
        else if (type == stateTransfer.GetTypeName())
        { // checked.
            stateTransfer.ParseFromString(data);
            HandleStateTransfer(remote, stateTransfer);
        }
        else if (type == startViewChange.GetTypeName())
        { // checked.
            startViewChange.ParseFromString(data);
            HandleStartViewChange(remote, startViewChange);
        }
        else if (type == doViewChange.GetTypeName())
        { // checked.
            doViewChange.ParseFromString(data);
            HandleDoViewChange(remote, doViewChange);
        }
        else if (type == startView.GetTypeName())
        { // checked.
            startView.ParseFromString(data);
            HandleStartView(remote, startView);
        }
        else if (type == recovery.GetTypeName())
        { // recovery is too rare, omit here.
            recovery.ParseFromString(data);
            HandleRecovery(remote, recovery);
        }
        else if (type == recoveryResponse.GetTypeName())
        { // recovery is too rare, omit here.
            recoveryResponse.ParseFromString(data);
            HandleRecoveryResponse(remote, recoveryResponse);
        }
        else
        {
            RPanic("Received unexpected message type in VR proto: %s",
                   type.c_str());
        }

        // asd123www: add logic here if we have handle view-change.
    }

    void VRReplica::HandleRequest(const TransportAddress &remote,
                                  const RequestMessage &msg)
    {
        // 后期可能要细看这个函数？相关注释很多 by：wbh
        viewstamp_t v;
        Latency_Start(&requestLatency);

        if (status != STATUS_NORMAL)
        { // don't handle request.
            RNotice("Ignoring request due to abnormal status");
            Latency_EndType(&requestLatency, 'i');
            return;
        }

        if (!AmLeader())
        { // only leader should handle request.
            RDebug("Ignoring request because I'm not the leader");
            Latency_EndType(&requestLatency, 'i');
            return;
        }

        // Save the client's address
        clientAddresses.erase(msg.req().clientid());
        clientAddresses.insert(
            std::pair<uint64_t, std::unique_ptr<TransportAddress>>(
                msg.req().clientid(),
                std::unique_ptr<TransportAddress>(remote.clone())));

        // Check the client table to see if this is a duplicate request
        auto kv = clientTable.find(msg.req().clientid());
        if (kv != clientTable.end())
        {
            const ClientTableEntry &entry = kv->second;
            if (msg.req().clientreqid() < entry.lastReqId)
            {
                RNotice("Ignoring stale request");
                Latency_EndType(&requestLatency, 's');
                return;
            }
            if (msg.req().clientreqid() == entry.lastReqId)
            {
                // This is a duplicate request. Resend the reply if we
                // have one. We might not have a reply to resend if we're
                // waiting for the other replicas; in that case, just
                // discard the request.
                if (entry.replied)
                {
                    RNotice("Received duplicate request; resending reply");
                    if (!(transport->SendMessage(this, remote,
                                                 entry.reply)))
                    {
                        RWarning("Failed to resend reply to client");
                    }
                    Latency_EndType(&requestLatency, 'r');
                    return;
                }
                else
                {
                    RNotice("Received duplicate request but no reply available; ignoring");
                    Latency_EndType(&requestLatency, 'd');
                    return;
                }
            }
        }

        // Update the client table
        UpdateClientTable(msg.req());

        // Leader Upcall
        bool replicate = false;
        string res;
        LeaderUpcall(lastCommitted, msg.req().op(), replicate, res); // set replicate and copy op() to res.
        ClientTableEntry &cte = clientTable[msg.req().clientid()];

        // Check whether this request should be committed to replicas
        if (!replicate)
        {
            RDebug("Executing request failed. Not committing to replicas");
            ReplyMessage reply;

            reply.set_reply(res);
            reply.set_view(0);
            reply.set_opnum(0);
            reply.set_clientreqid(msg.req().clientreqid());
            cte.replied = true;
            cte.reply = reply;
            transport->SendMessage(this, remote, reply);
            Latency_EndType(&requestLatency, 'f');
        }
        else
        { // here is the normal execution path.
            Request request;
            request.set_op(res);
            request.set_clientid(msg.req().clientid());
            request.set_clientreqid(msg.req().clientreqid());

            /* Assign it an opnum */ // increasing by one.
            ++this->lastOp;
            v.view = this->view;
            v.opnum = this->lastOp;

            RDebug("Received REQUEST, assigning " FMT_VIEWSTAMP, VA_VIEWSTAMP(v));

            /* Add the request to my log */
            log.Append(v, request, LOG_STATE_PREPARED); // state of this entry in log is PREPARED.

            if (batchComplete ||
                (lastOp - lastBatchEnd + 1 > (unsigned int)batchSize))
            {
                CloseBatch();
            }
            else
            {
                RDebug("Keeping in batch");
                if (!closeBatchTimeout->Active())
                {
                    closeBatchTimeout->Start();
                }
            }

            nullCommitTimeout->Reset(); // asd123www: 这合理吗？我这次又不一定发信息....
            Latency_End(&requestLatency);
        }
    }

    // asd123www: ?????only client send this message, rare case.
    void VRReplica::HandleUnloggedRequest(const TransportAddress &remote, const UnloggedRequestMessage &msg)
    {
        if (status != STATUS_NORMAL)
        {
            // Not clear if we should ignore this or just let the request
            // go ahead, but this seems reasonable.
            RNotice("Ignoring unlogged request due to abnormal status");
            return;
        }

        UnloggedReplyMessage reply;

        Debug("Received unlogged request %s", (char *)msg.req().op().c_str());

        ExecuteUnlogged(msg.req(), reply);

        if (!(transport->SendMessage(this, remote, reply)))
            Warning("Failed to send reply message");
    }

    void VRReplica::HandlePrepare_Kernel(const proto::PrepareMessage &msg)
    {
        // 新增HandlePrepare_Kernel函数
        RDebug("Received PREPARE <" FMT_VIEW "," FMT_OPNUM "-" FMT_OPNUM ">",
               msg.view(), msg.batchstart(), msg.opnum());
        // 确保当前节点处于正常状态
        ASSERT(this->status == STATUS_NORMAL);
        ASSERT(msg.view() == this->view);
        ASSERT(msg.batchstart() <= msg.opnum());
        ASSERT_EQ(msg.opnum() - msg.batchstart() + 1, (unsigned)msg.request_size());

        /* Add operations to the log */
        opnum_t op = msg.batchstart() - 1;
        for (auto &req : msg.request())
        {
            // 根据PREPARE消息中的请求信息更新日志，并更新最后操作编号和客户端表
            op++;
            if (op <= lastOp)
                continue;
            this->lastOp++;
            log.Append(viewstamp_t(msg.view(), op), req, LOG_STATE_PREPARED);
            UpdateClientTable(req);
        }
        ASSERT(op == msg.opnum());
    }

    void VRReplica::HandlePrepare(const TransportAddress &remote, const PrepareMessage &msg)
    {
        RDebug("Received PREPARE <" FMT_VIEW "," FMT_OPNUM "-" FMT_OPNUM ">",
               msg.view(), msg.batchstart(), msg.opnum());

        if (this->status != STATUS_NORMAL)
        { // no interaction.
            RDebug("Ignoring PREPARE due to abnormal status");
            return;
        }

        if (msg.view() < this->view)
        { // hear a stale  message, we shouldn't respond to that.
            RDebug("Ignoring PREPARE due to stale view");
            return;
        }

        if (msg.view() > this->view)
        { // trigger view change! we laid behind.
            RequestStateTransfer();
            pendingPrepares.push_back(std::pair<TransportAddress *, PrepareMessage>(remote.clone(), msg));
            return;
        }

        if (AmLeader())
        { // leader shouldn't receive this message.
            RPanic("Unexpected PREPARE: I'm the leader of this view");
        }

        // sanity checks.
        ASSERT(msg.batchstart() <= msg.opnum());
        ASSERT_EQ(msg.opnum() - msg.batchstart() + 1, (unsigned)msg.request_size());

        viewChangeTimeout->Reset(); // asd123www: how to maintain this in XDP? maybe pulling from kernel if this is triggered.

        if (msg.opnum() <= this->lastOp)
        { // we have received...
            RDebug("Ignoring PREPARE; already prepared that operation");
            // Resend the prepareOK message
            PrepareOKMessage reply;
            reply.set_view(msg.view());
            reply.set_opnum(msg.opnum());
            reply.set_replicaidx(myIdx);

            uint32_t my_data[3] = {(uint32_t)msg.view(), (uint32_t)msg.opnum(), (uint32_t)myIdx};
            if (!(transport->SendMessageToReplica(this,
                                                  configuration.GetLeaderIndex(view),
                                                  reply, my_data)))
            {
                RWarning("Failed to send PrepareOK message to leader");
            }
            return;
        }

        /* asd123www: From <Viewstamped Replication Revisited>:
            a backup won’t accept a prepare with op-number n
            until it has entries for all earlier requests in its log. */
        if (msg.batchstart() > this->lastOp + 1)
        {
            RequestStateTransfer();
            pendingPrepares.push_back(std::pair<TransportAddress *, PrepareMessage>(remote.clone(), msg));
            return;
        }

        /* Add operations to the log */
        opnum_t op = msg.batchstart() - 1;
        for (auto &req : msg.request())
        {
            op++;
            if (op <= lastOp)
                continue;
            this->lastOp++;
            log.Append(viewstamp_t(msg.view(), op), req, LOG_STATE_PREPARED);
            UpdateClientTable(req);
        }
        ASSERT(op == msg.opnum());

        /* Build reply and send it to the leader */
        PrepareOKMessage reply;
        reply.set_view(msg.view());
        reply.set_opnum(msg.opnum());
        reply.set_replicaidx(myIdx);

        uint32_t my_data[3] = {(uint32_t)msg.view(), (uint32_t)msg.opnum(), (uint32_t)myIdx};
        if (!(transport->SendMessageToReplica(this,
                                              configuration.GetLeaderIndex(view),
                                              reply, my_data)))
        {
            RWarning("Failed to send PrepareOK message to leader");
        }
    }

    void VRReplica::HandlePrepareOK(const TransportAddress &remote, const PrepareOKMessage &msg)
    {
        RDebug("Received PREPAREOK <" FMT_VIEW ", " FMT_OPNUM "> from replica %d",
               msg.view(), msg.opnum(), msg.replicaidx());

        if (this->status != STATUS_NORMAL)
        {
            RDebug("Ignoring PREPAREOK due to abnormal status");
            return;
        }

        if (msg.view() < this->view)
        {
            RDebug("Ignoring PREPAREOK due to stale view");
            return;
        }

        if (msg.view() > this->view)
        {
            RequestStateTransfer();
            return;
        }

        if (!AmLeader())
        {
            RWarning("Ignoring PREPAREOK because I'm not the leader");
            return;
        }

        viewstamp_t vs = {msg.view(), msg.opnum()};
#ifndef FAST_QUORUM_PRUNE
        if (auto msgs = (prepareOKQuorum.AddAndCheckForQuorum(vs, msg.replicaidx(), msg)))
        {
            // 如果宏定义了FAST_QUORUM_PRUNE，就不需要if的操作了，if和宏判断的是同一件事情。 by：wbh
#endif

            /*
             * We have a quorum of PrepareOK messages for this
             * opnumber. Execute it and all previous operations.
             *
             * (Note that we might have already executed it. That's fine,
             * we just won't do anything.)
             *
             * This also notifies the client of the result.
             */
            CommitUpTo(msg.opnum()); // leader then knows these messages have been committed, then we can apply to the state-machine.

#ifndef FAST_QUORUM_PRUNE
            if (msgs->size() >= (unsigned)configuration.QuorumSize())
            { // not the critical point.. e.g. Quorum() - 1
                return;
            }
#endif
            /*
             * Send COMMIT message to the other replicas.
             *
             * This can be done asynchronously, so it really ought to be
             * piggybacked on the next PREPARE or something.
             */
            CommitMessage cm;
            cm.set_view(this->view);
            cm.set_opnum(this->lastCommitted);

            if (!(transport->SendMessageToAll(this, cm
#ifdef TC_BROADCAST
                                              ,
                                              sgn_bits
#endif
                                              )))
            {
                RWarning("Failed to send COMMIT message to all replicas");
            }

            nullCommitTimeout->Reset(); // we will broadcast a message, then this can be cancelled.

            // asd123www: what is adaptive batching????
            // XXX Adaptive batching -- make this configurable
            if (lastBatchEnd == msg.opnum())
            {
                batchComplete = true;
                if (lastOp > lastBatchEnd)
                {
                    CloseBatch();
                }
            }
#ifndef FAST_QUORUM_PRUNE
        }
#endif
    }

    void VRReplica::HandleCommit(const TransportAddress &remote, const CommitMessage &msg)
    {
        /**
         * 处理提交(Commit)消息。
         *
         * @param remote 表示发送方的TransportAddress。
         * @param msg 表示接收到的提交(Commit)消息。
         *
         * 此函数检查提交消息的有效性，并根据当前的状态和消息内容更新系统状态。
         * 主要步骤包括：
         * 1. 日志记录接收到的提交消息。
         * 2. 检查系统当前状态是否正常，如果不正常则忽略该提交消息。
         * 3. 检查提交消息的视图(view)是否过时或超前，如果过时或超前则忽略该消息。
         * 4. 如果当前节点是领导者，则认为这是一个异常情况，记录错误并返回。
         * 5. 重置视图变更超时计时器。
         * 6. 检查提交的消息操作号是否已经提交，如果是，则忽略该消息。
         * 7. 如果消息的操作号大于当前节点记录的最后一个操作号，则请求状态转移。
         * 8. 最后，提交到消息的操作号。
         */
        // 记录接收到的COMMIT消息的日志
        RDebug("Received COMMIT " FMT_VIEWSTAMP, msg.view(), msg.opnum());

        // 如果当前状态非正常，则忽略此COMMIT
        if (this->status != STATUS_NORMAL)
        {
            RDebug("Ignoring COMMIT due to abnormal status");
            return;
        }

        // 检查消息的视图是否过时
        if (msg.view() < this->view)
        {
            RDebug("Ignoring COMMIT due to stale view");
            return;
        }

        // 如果消息的视图较新，请求状态转移
        if (msg.view() > this->view)
        {
            RequestStateTransfer();
            return;
        }

        // 如果当前节点是领导者，则记录错误并返回
        if (AmLeader())
        {
            RPanic("Unexpected COMMIT: I'm the leader of this view");
        }

        // 重置视图变更的超时计时器
        viewChangeTimeout->Reset();

        // 检查消息的操作号是否已经提交
        if (msg.opnum() <= this->lastCommitted)
        {
            RDebug("Ignoring COMMIT; already committed that operation");
            return;
        }

        // 如果操作号大于当前记录的最后一个操作号，请求状态转移
        if (msg.opnum() > this->lastOp)
        {
            RequestStateTransfer();
            return;
        }

        // 提交到消息的操作号
        CommitUpTo(msg.opnum());
    }

    void VRReplica::HandleRequestStateTransfer(const TransportAddress &remote, const RequestStateTransferMessage &msg)
    {
        /**
         * 处理状态传输请求的函数。
         *
         * @param remote 源地址，表示请求状态传输的客户端。
         * @param msg 包含状态传输请求详细信息的消息对象。
         *
         * 此函数首先检查当前系统状态是否正常，然后根据请求的消息视图和操作号来决定是进行状态传输，
         * 还是仅仅回复当前系统状态。它会将当前系统的状态封装在响应消息中，并发送给请求者。
         */
        // 记录接收到的状态传输请求
        RDebug("Received REQUESTSTATETRANSFER " FMT_VIEWSTAMP,
               msg.view(), msg.opnum());

        // 如果系统状态异常，则忽略此请求
        if (status != STATUS_NORMAL)
        {
            RDebug("Ignoring REQUESTSTATETRANSFER due to abnormal status");
            return;
        }

        // 如果请求的视图大于当前视图，则发起状态传输
        if (msg.view() > view)
        {
            RequestStateTransfer();
            return;
        }

        // 记录正在发送的状态传输信息
        RNotice("Sending state transfer from " FMT_VIEWSTAMP " to " FMT_VIEWSTAMP,
                msg.view(), msg.opnum(), view, lastCommitted);

        // 准备状态传输的回复消息
        StateTransferMessage reply;
        reply.set_view(view);
        reply.set_opnum(lastCommitted);

        // 将日志状态从指定操作号开始打包到消息中
        log.Dump(msg.opnum() + 1, reply.mutable_entries());

        // 向请求者发送包含当前状态的回复消息
        transport->SendMessage(this, remote, reply);
    }

    // update our state.
    void VRReplica::HandleStateTransfer(const TransportAddress &remote, const StateTransferMessage &msg)
    {
        RDebug("Received STATETRANSFER " FMT_VIEWSTAMP, msg.view(), msg.opnum());

        if (msg.view() < view)
        {
            RWarning("Ignoring state transfer for older view");
            return;
        }

        opnum_t oldLastOp = lastOp;

        /* Install the new log entries */
        for (auto newEntry : msg.entries())
        {
            if (newEntry.opnum() <= lastCommitted)
            {
                // Already committed this operation; nothing to be done.
#if PARANOID
                const LogEntry *entry = log.Find(newEntry.opnum());
                ASSERT(entry->viewstamp.opnum == newEntry.opnum());
                ASSERT(entry->viewstamp.view == newEntry.view());
//          ASSERT(entry->request == newEntry.request());
#endif
            }
            else if (newEntry.opnum() <= lastOp)
            {
                // We already have an entry with this opnum, but maybe
                // it's from an older view?
                const LogEntry *entry = log.Find(newEntry.opnum());
                ASSERT(entry->viewstamp.opnum == newEntry.opnum());
                ASSERT(entry->viewstamp.view <= newEntry.view());

                if (entry->viewstamp.view == newEntry.view())
                {
                    // We already have this operation in our log.
                    ASSERT(entry->state == LOG_STATE_PREPARED);
#if PARANOID
//              ASSERT(entry->request == newEntry.request());
#endif
                }
                else
                {
                    // Our operation was from an older view, so obviously
                    // it didn't survive a view change. Throw out any
                    // later log entries and replace with this one.
                    ASSERT(entry->state != LOG_STATE_COMMITTED);
                    log.RemoveAfter(newEntry.opnum());
                    lastOp = newEntry.opnum();
                    oldLastOp = lastOp;

                    viewstamp_t vs = {newEntry.view(), newEntry.opnum()};
                    log.Append(vs, newEntry.request(), LOG_STATE_PREPARED);
                }
            }
            else
            {
                // This is a new operation to us. Add it to the log.
                ASSERT(newEntry.opnum() == lastOp + 1);

                lastOp++;
                viewstamp_t vs = {newEntry.view(), newEntry.opnum()};
                log.Append(vs, newEntry.request(), LOG_STATE_PREPARED);
            }
        }

        if (msg.view() > view)
        {
            EnterView(msg.view());
        }

        /* Execute committed operations */
        ASSERT(msg.opnum() <= lastOp);
        CommitUpTo(msg.opnum());
        SendPrepareOKs(oldLastOp);

        // Process pending prepares
        std::list<std::pair<TransportAddress *, PrepareMessage>> pending = pendingPrepares;
        pendingPrepares.clear();
        for (auto &msgpair : pendingPrepares)
        {
            RDebug("Processing pending prepare message");
            HandlePrepare(*msgpair.first, msgpair.second);
            delete msgpair.first;
        }
    }

    void VRReplica::HandleStartViewChange(const TransportAddress &remote, const StartViewChangeMessage &msg)
    {
        RDebug("Received STARTVIEWCHANGE " FMT_VIEW " from replica %d",
               msg.view(), msg.replicaidx());

        if (msg.view() < view)
        {
            RDebug("Ignoring STARTVIEWCHANGE for older view");
            return;
        }

        if ((msg.view() == view) && (status != STATUS_VIEW_CHANGE))
        {
            RDebug("Ignoring STARTVIEWCHANGE for current view");
            return;
        }

        if ((status != STATUS_VIEW_CHANGE) || (msg.view() > view))
        {
            RWarning("Received StartViewChange for view " FMT_VIEW
                     "from replica %d",
                     msg.view(), msg.replicaidx());
            StartViewChange(msg.view());
        }

        ASSERT(msg.view() == view);

        // if reach quorum, then broadcast do view change message.
        if (auto msgs =
                startViewChangeQuorum.AddAndCheckForQuorum(msg.view(),
                                                           msg.replicaidx(),
                                                           msg))
        {
            int leader = configuration.GetLeaderIndex(view);
            // Don't try to send a DoViewChange message to ourselves
            if (leader != myIdx)
            {
                DoViewChangeMessage dvc;
                dvc.set_view(view);
                dvc.set_lastnormalview(log.LastViewstamp().view);
                dvc.set_lastop(lastOp);
                dvc.set_lastcommitted(lastCommitted);
                dvc.set_replicaidx(myIdx);

                // Figure out how much of the log to include
                opnum_t minCommitted = std::min_element(
                                           msgs->begin(), msgs->end(),
                                           [](decltype(*msgs->begin()) a,
                                              decltype(*msgs->begin()) b)
                                           {
                                               return a.second.lastcommitted() < b.second.lastcommitted();
                                           })
                                           ->second.lastcommitted();
                minCommitted = std::min(minCommitted, lastCommitted);

                log.Dump(minCommitted,
                         dvc.mutable_entries());

                if (!(transport->SendMessageToReplica(this, leader, dvc)))
                {
                    RWarning("Failed to send DoViewChange message to leader of new view");
                }
            }
        }
    }

    void VRReplica::HandleDoViewChange(const TransportAddress &remote, const DoViewChangeMessage &msg)
    {
        RDebug("Received DOVIEWCHANGE " FMT_VIEW " from replica %d, "
               "lastnormalview=" FMT_VIEW " op=" FMT_OPNUM " committed=" FMT_OPNUM,
               msg.view(), msg.replicaidx(),
               msg.lastnormalview(), msg.lastop(), msg.lastcommitted());

        if (msg.view() < view)
        {
            RDebug("Ignoring DOVIEWCHANGE for older view");
            return;
        }

        if ((msg.view() == view) && (status != STATUS_VIEW_CHANGE))
        {
            RDebug("Ignoring DOVIEWCHANGE for current view");
            return;
        }

        if ((status != STATUS_VIEW_CHANGE) || (msg.view() > view))
        {
            // It's superfluous to send the StartViewChange messages here,
            // but harmless...
            RWarning("Received DoViewChange for view " FMT_VIEW
                     "from replica %d",
                     msg.view(), msg.replicaidx());
            StartViewChange(msg.view());
        }

        ASSERT(configuration.GetLeaderIndex(msg.view()) == myIdx);

        auto msgs = doViewChangeQuorum.AddAndCheckForQuorum(msg.view(),
                                                            msg.replicaidx(),
                                                            msg);
        if (msgs != NULL)
        {
            // Find the response with the most up to date log, i.e. the
            // one with the latest viewstamp
            view_t latestView = log.LastViewstamp().view;
            opnum_t latestOp = log.LastViewstamp().opnum;
            DoViewChangeMessage *latestMsg = NULL;

            for (auto kv : *msgs)
            {
                DoViewChangeMessage &x = kv.second;
                if ((x.lastnormalview() > latestView) ||
                    (((x.lastnormalview() == latestView) &&
                      (x.lastop() > latestOp))))
                {
                    latestView = x.lastnormalview();
                    latestOp = x.lastop();
                    latestMsg = &x;
                }
            }

            // Install the new log. We might not need to do this, if our
            // log was the most current one.
            if (latestMsg != NULL)
            {
                RDebug("Selected log from replica %d with lastop=" FMT_OPNUM,
                       latestMsg->replicaidx(), latestMsg->lastop());
                if (latestMsg->entries_size() == 0)
                {
                    // There weren't actually any entries in the
                    // log. That should only happen in the corner case
                    // that everyone already had the entire log, maybe
                    // because it actually is empty.
                    ASSERT(lastCommitted == msg.lastcommitted());
                    ASSERT(msg.lastop() == msg.lastcommitted());
                }
                else
                {
                    if (latestMsg->entries(0).opnum() > lastCommitted + 1)
                    {
                        RPanic("Received log that didn't include enough entries to install it");
                    }

                    log.RemoveAfter(latestMsg->lastop() + 1);
                    log.Install(latestMsg->entries().begin(),
                                latestMsg->entries().end());
                }
            }
            else
            {
                RDebug("My log is most current, lastnormalview=" FMT_VIEW " lastop=" FMT_OPNUM,
                       log.LastViewstamp().view, lastOp);
            }

            // How much of the log should we include when we send the
            // STARTVIEW message? Start from the lowest committed opnum of
            // any of the STARTVIEWCHANGE or DOVIEWCHANGE messages we got.
            //
            // We need to compute this before we enter the new view
            // because the saved messages will go away.
            auto svcs = startViewChangeQuorum.GetMessages(view);
            opnum_t minCommittedSVC = std::min_element(
                                          svcs.begin(), svcs.end(),
                                          [](decltype(*svcs.begin()) a,
                                             decltype(*svcs.begin()) b)
                                          {
                                              return a.second.lastcommitted() < b.second.lastcommitted();
                                          })
                                          ->second.lastcommitted();
            opnum_t minCommittedDVC = std::min_element(
                                          msgs->begin(), msgs->end(),
                                          [](decltype(*msgs->begin()) a,
                                             decltype(*msgs->begin()) b)
                                          {
                                              return a.second.lastcommitted() < b.second.lastcommitted();
                                          })
                                          ->second.lastcommitted();
            opnum_t minCommitted = std::min(minCommittedSVC, minCommittedDVC);
            minCommitted = std::min(minCommitted, lastCommitted);

            EnterView(msg.view());

            ASSERT(AmLeader());

            lastOp = latestOp;
            if (latestMsg != NULL)
            {
                CommitUpTo(latestMsg->lastcommitted());
            }

            // Send a STARTVIEW message with the new log
            StartViewMessage sv;
            sv.set_view(view);
            sv.set_lastop(lastOp);
            sv.set_lastcommitted(lastCommitted);

            log.Dump(minCommitted, sv.mutable_entries());

            if (!(transport->SendMessageToAll(this, sv)))
            {
                RWarning("Failed to send StartView message to all replicas");
            }
        }
    }

    void
    VRReplica::HandleStartView(const TransportAddress &remote,
                               const StartViewMessage &msg)
    {
        RDebug("Received STARTVIEW " FMT_VIEW
               " op=" FMT_OPNUM " committed=" FMT_OPNUM " entries=%d",
               msg.view(), msg.lastop(), msg.lastcommitted(), msg.entries_size());
        RDebug("Currently in view " FMT_VIEW " op " FMT_OPNUM " committed " FMT_OPNUM,
               view, lastOp, lastCommitted);

        if (msg.view() < view)
        {
            RWarning("Ignoring STARTVIEW for older view");
            return;
        }

        if ((msg.view() == view) && (status != STATUS_VIEW_CHANGE))
        {
            RWarning("Ignoring STARTVIEW for current view");
            return;
        }

        ASSERT(configuration.GetLeaderIndex(msg.view()) != myIdx);

        if (msg.entries_size() == 0)
        {
            ASSERT(msg.lastcommitted() == lastCommitted);
            ASSERT(msg.lastop() == msg.lastcommitted());
        }
        else
        {
            if (msg.entries(0).opnum() > lastCommitted + 1)
            {
                RPanic("Not enough entries in STARTVIEW message to install new log");
            }

            // Install the new log
            log.RemoveAfter(msg.lastop() + 1);
            log.Install(msg.entries().begin(),
                        msg.entries().end());
        }

        EnterView(msg.view());
        opnum_t oldLastOp = lastOp;
        lastOp = msg.lastop();

        ASSERT(!AmLeader());

        CommitUpTo(msg.lastcommitted());
        SendPrepareOKs(oldLastOp);
    }

    void
    VRReplica::HandleRecovery(const TransportAddress &remote,
                              const RecoveryMessage &msg)
    {
        RDebug("Received RECOVERY from replica %d", msg.replicaidx());

        if (status != STATUS_NORMAL)
        {
            RDebug("Ignoring RECOVERY due to abnormal status");
            return;
        }

        RecoveryResponseMessage reply;
        reply.set_replicaidx(myIdx);
        reply.set_view(view);
        reply.set_nonce(msg.nonce());
        if (AmLeader())
        {
            reply.set_lastcommitted(lastCommitted);
            reply.set_lastop(lastOp);
            log.Dump(0, reply.mutable_entries());
        }

        if (!(transport->SendMessage(this, remote, reply)))
        {
            RWarning("Failed to send recovery response");
        }
        return;
    }

    void
    VRReplica::HandleRecoveryResponse(const TransportAddress &remote,
                                      const RecoveryResponseMessage &msg)
    {
        RDebug("Received RECOVERYRESPONSE from replica %d",
               msg.replicaidx());

        if (status != STATUS_RECOVERING)
        {
            RDebug("Ignoring RECOVERYRESPONSE because we're not recovering");
            return;
        }

        if (msg.nonce() != recoveryNonce)
        {
            RNotice("Ignoring recovery response because nonce didn't match");
            return;
        }

        auto msgs = recoveryResponseQuorum.AddAndCheckForQuorum(msg.nonce(),
                                                                msg.replicaidx(),
                                                                msg);
        if (msgs != NULL)
        {
            view_t highestView = 0;
            for (const auto &kv : *msgs)
            {
                if (kv.second.view() > highestView)
                {
                    highestView = kv.second.view();
                }
            }

            int leader = configuration.GetLeaderIndex(highestView);
            ASSERT(leader != myIdx);
            auto leaderResponse = msgs->find(leader);
            if ((leaderResponse == msgs->end()) ||
                (leaderResponse->second.view() != highestView))
            {
                RDebug("Have quorum of RECOVERYRESPONSE messages, "
                       "but still need to wait for one from the leader");
                return;
            }

            Notice("Recovery completed");

            log.Install(leaderResponse->second.entries().begin(),
                        leaderResponse->second.entries().end());
            EnterView(leaderResponse->second.view());
            lastOp = leaderResponse->second.lastop();
            CommitUpTo(leaderResponse->second.lastcommitted());
        }
    }

} // namespace specpaxos::vr
} // namespace specpaxos
